#ifndef GMesh_H_
#define GMesh_H_
#include <string> 
#include <cmath>
#include <cstdlib>
#include <iostream>
#include <fstream>
#include <set>
#include <string> 

#include "error.hpp"
#include "throwassert.hpp"

using namespace std;

namespace GMesh3D  
{
//Quelques definition de base
template<class T> inline T Min (const T &a,const T &b){ return a < b ? a : b; }
template<class T> inline T Max (const T &a,const T &b,const T &c){ return Max(Max(a,b),c); }
template<class T> inline T Max (const T &a,const T &b){ return a > b ? a : b; }
template<class T> inline T Min (const T &a,const T &b,const T &c){ return Min(Min(a,b),c); }
template<class T> inline T Abs (const T &a){return a <0 ? -a : a; }


// R comme reel
typedef double R;

//---------------------------------------------------------------------------------------
//                                       Classe R3
//---------------------------------------------------------------------------------------
class R3 
{   public:
    // coordonnees x,y et z d'un point 
    R  x,y,z;
    // initialisation a 0    
    R3 () :x(0),y(0),z(0) {}
    // affectation de valeurs    
    R3 (R a,R b,R c):x(a),y(b),z(c)  {}
    // distance    
    R3 (R3 A,R3 B) :x(B.x-A.x),y(B.y-A.y),z(B.z-A.z)  {}
    // surcharge des operateurs    
    R3   operator +   (R3 P) const { return R3(x+P.x,y+P.y,z+P.z); }
    R3   operator -   (R3 P) const { return R3(x-P.x,y-P.y,z-P.z); }    
    R3   operator +=  (R3 P)       { x += P.x; y += P.y;z += P.z; return *this; }    
    R3   operator -=  (R3 P)       { x -= P.x; y -= P.y;z -= P.z; return *this; }
    int  operator ==  (R3 P)       { int E=0; if (x==P.x && y==P.y && z==P.z){E =1;};return E; } 
    R3   operator -()        const { return R3(-x,-y,-z); }
    R3   operator +()        const { return *this; }
    // le produit scalaire	 
    R    operator ,   (R3 P) const { return  x*P.x + y*P.y + z*P.z; } 
    // le produit vectoriel	 
    R3   operator ^   (R3 P) const { return  R3(y*P.z-z*P.y ,P.x*z-x*P.z, x*P.y-y*P.x); }
    R3   operator *   (R  c) const { return R3(x*c,y*c,z*c); }
    R3   operator /   (R  c) const { return R3(x/c,y/c,z/c); }
    R3   operator *=  (R  c)       { x *= c; y *= c; z *= c; return *this; }
    R3   operator /=  (R  c)       { x /= c; y /= c; z /= c; return *this; }       
    R  & operator []  (int i)      { return (&x)[i];}

    // affichage 
    friend ostream& operator <<(ostream& f, const R3 & P )
    { f << P.x << ' ' << P.y << ' ' << P.z   ; return f; }
    // ecriture
    friend istream& operator >>(istream& f,  R3 & P)
    { f >>  P.x >>  P.y >>  P.z  ; return f; }
};
// Definition de fonction utile
// norme au carre
inline R Norme2_2(const R3 & A)   { return (A,A); }
// racine de la norme
inline R Norme2  (const R3 & A)   { return sqrt((A,A)); }
// norme infini
inline R Norme_infty(const R3 & A){ return Max(Abs(A.x),Abs(A.y),Abs(A.z)); }

//-------------------------------------------------------------------------------------
//                                       Classe label
//-------------------------------------------------------------------------------------
class Label
{
public: 
  int lab;
  //constructeur
  Label(int r=0):lab(r){}
  int  operator !()                 const { return !lab;} 
  bool operator < (const Label & r) const { return lab < r.lab; } 
  bool operator ==(const Label & r) const { return lab == r.lab;} 
  bool operator > (const Label & r) const { return lab > r.lab; } 
};  
// affichage
inline ostream& operator <<( ostream& f, const Label & r ) { f <<  r.lab ; return f; }
  // ecriture
inline istream& operator >>( istream& f, Label & r  )      { f >>  r.lab ; return f; }

//--------------------------------------------------------------------------------------
//                                    Classe vertex = Sommet
//--------------------------------------------------------------------------------------
class Vertex : public R3, public Label 
{
  friend class Mesh;
private:
   R3 * normal; //pointeur sur la normale exterieure pour filtrer les points de depart 
public:

  int color;
  Vertex() : R3(), Label(), normal(0) {};
  Vertex( R3 P, int r=0 ) : R3(P), Label(r), normal(0) {}  
  void color_efface()
  { color = 0; }
  void colorise()
  { color = 1; } 
  void color_pas_de_chance()
  { color = -1;}
  bool ninside( const R3 & P ) const { return normal? (R3(*this,P),*normal)<=0 : true; }
  void SetNormal( R3 *&n, const R3 & N )
  { if (normal)
    { 
      R3 NN=*normal+N; 
      *normal= NN/Norme2(NN); 
    }
    else
      *(normal=n++)=N;
  }
  R3 Ne() const { return normal ? *normal : R3(0,0,0); }
};
// affichage
inline ostream& operator << ( ostream& f, const Vertex & v )
       { f << (R3) v << ' ' << (Label) v   ; return f; }
// ecriture  
inline istream& operator >> ( istream& f,  Vertex & v )
       { f >> (R3&) v >> (Label&) v ; return f; }


//-------------------------------------------------------------------------------------
//                                    classe Tetraedre
//-------------------------------------------------------------------------------------
class Tetraedre: public Label  
{
public:
  Vertex *vertices[4]; 
  R volume;
  int Neigh[4];
  int color;

  // construteur par default
  Tetraedre(){};       
  Tetraedre( Vertex * v0, int i0, int i1, int i2,int i3, int l, R vol=0.0 ) : Label(l) 
  { R3 A=*(vertices[0]=v0+i0);
    R3 B=*(vertices[1]=v0+i1);
    R3 C=*(vertices[2]=v0+i2); 
    R3 D=*(vertices[3]=v0+i3); 		 
    volume = vol==0 ? (((B-A)^(C-A)),(D-A))/6. : vol;
    throwassert(volume>0);}
  void init( Vertex * v0, int i0, int i1, int i2,int i3, int l, R vol=0.0 )
  { lab=l; 
    R3 A=*(vertices[0]=v0+i0);
    R3 B=*(vertices[1]=v0+i1);
    R3 C=*(vertices[2]=v0+i2); 
    R3 D=*(vertices[3]=v0+i3); 
    volume = vol==0 ? (((B-A)^(C-A)),(D-A))/6. : vol;
    throwassert(volume>0);
    // initialisation de l'adjacence
    for (int k=0;k<4;k++)    Neigh[k] = -1;
  }
  void color_efface()
  { color = 0; }
  void colorise()
  { color = 1; }  
  void voisinage(int k,int j)
  {
    Neigh[j] = k;
  }
  Vertex &   operator[](int i) const {  return *vertices[i]; }        
  Vertex * & operator()(int i)       {  return  vertices[i]; } 
  R3 H(int i)  const  
  {
     int j1=(i+1)%4;
     int j2=(i+2)%4;
     int j3=(i+3)%4;
     R3 ortho = (*vertices[j2]-*vertices[j1])^(*vertices[j3]-*vertices[j1]);
     R3 gradient = ortho/((*vertices[i]-*vertices[j1]),ortho);
     return gradient; 
  }      
private:
  Tetraedre(const Tetraedre &); 
  void operator=(const Tetraedre &);             
};
//-----------------------------------------------------------------------------------
//                         Classe Mesh = generation d'un maillage 
//-----------------------------------------------------------------------------------
class Mesh
{
  public:
    int  dim;    // dimension 
    int  nt;     // nombre de tetraedres
    int  nv;     // nombre de sommets
    R    volume; //total du maillage

    Vertex         * vertices;      // tableau des vertices du maillage
    Tetraedre      * tetraedres;    // tableau des tetraedres du maillage
  

    // Le constructeur
    // lecture dans les fichiers
    Mesh( const char * maillage ) 
    { read(maillage);   Buildadjacence();} 
    // Le destructeur
   ~Mesh();
    void Buildadjacence();
    /*
     * calcule le max des longueurs des aretes
     */
    R    taille_caracteristique(void);
    // Quelques liens utiles 
    // le triangle i du maillage
    Tetraedre & operator[](int i)   const { throwassert(i>=0 && i<nt); return tetraedres[i]; }
    // le vertex i du maillage
    Vertex    & operator()(int i)   const { throwassert(i>=0 && i<nv); return   vertices[i]; }
    int number(const Tetraedre & t) const { return &t - tetraedres;}
    int number(const Tetraedre * t) const { return  t - tetraedres;}  
    int number(const Vertex & v)   const  { return &v - vertices;}
    int number(const Vertex * v)   const  { return  v - vertices;}

    int operator()(const Tetraedre & t) const { return &t - tetraedres;}
    int operator()(const Tetraedre * t) const { return  t - tetraedres;}
    int operator()(const Vertex & v)   const { return &v - vertices;}
    int operator()(const Vertex * v)   const { return  v - vertices;}
    // numero du vertex j du triangle it
    int operator()(int it, int j)       const {return number(tetraedres[it][j]);}
  private:
    void read(const char * maillage);
};



#define RESULT_TYPE_SCALAIRE 1
#define RESULT_TYPE_VECTORIEL 3

class Result
{
  public:
    int dim;
    int type;
    int nv;
    int typedonnees;

    R * scalaires;
    R3 * velocity;
    
    Result( const char * solution,  Mesh & Th) { read(solution,Th);} 
    Result( Mesh & Th2,int letype,int letypedonnees);
    ~Result(){}
    
    R operator[](int i)   const { throwassert(scalaires && i>=0 && i<nv); return   scalaires[i]; }
    R3 operator()(int i)   const { throwassert(velocity && i>=0 && i<nv); return   velocity[i]; }

  private:
    void read(const char * solution,Mesh & Th);
};

class Pile
{
public :
      Pile(int nv) : sp(0)
      {
         K = new int [nv];
         nb_max = nv;
         for (int i=0;i<nv;i++)
         {  K[i] = 0;  }   
      } 
      ~Pile() { delete [] K;}
      void push(int x) { K[sp++] = x; assert(sp<=nb_max); }
      int top()  { return K[sp-1]; }
      int pop()  { return K[--sp]; }
      int kpile(){ return sp;} 
private:
   int nb_max;
   int sp;
   int *K;
};

}

#endif

